using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using SimpleJSON;
using System.Linq;

namespace ClockwiseSilver
{
	public class SilverRoper : MVRScript
	{
		private Atom receivingAtom;
		private JSONStorableStringChooser atomJSON;
		private JSONStorable receiver;
		private JSONStorableStringChooser receiverJSON;
		private Atom receivingAtom2;
		private JSONStorableStringChooser atom2JSON;
		private JSONStorable receiver2;
		private JSONStorableStringChooser receiver2JSON;
		private JSONStorableBool isAttached;
		private JSONStorableBool disableRot;
		private JSONStorableFloat ropeLength;
		private JSONStorableFloat ropeRenderExtend;
		private JSONStorableFloat ropeRadius;
		private JSONStorableFloat caOffsetX;
		private JSONStorableFloat caOffsetY;
		private JSONStorableFloat caOffsetZ;
		private JSONStorableFloat caOffset2X;
		private JSONStorableFloat caOffset2Y;
		private JSONStorableFloat caOffset2Z;
		private JSONStorableColor ropeColor;

		private UIDynamicColorPicker colorPicker;		
		
		private Dictionary<string, FreeControllerV3> controllers = new Dictionary<string, FreeControllerV3>();
		private Dictionary<string, FreeControllerV3> controllers2 = new Dictionary<string, FreeControllerV3>();
		private FreeControllerV3 targetController;
		private Rigidbody targetBody;
		private FreeControllerV3.RotationState rState;
	    private FreeControllerV3.PositionState pState;
		private FreeControllerV3 targetController2;
		private Rigidbody targetBody2;
		private FreeControllerV3.RotationState rState2;
	    private FreeControllerV3.PositionState pState2;
		private ConfigurableJoint cj;
		private Vector3 startPos;
		private Quaternion startRot;
		private Vector3 startPos2;
		private Quaternion startRot2;
		private GameObject rope;
		private Material ropeMat;
		private Vector3 fixedOffset = Vector3.zero;
		private Vector3 fixedOffset2 = Vector3.zero;
		
		private float ropeScale = 0.01f;
		private float ropeLengthMulti = 1f;
		private string lastReceiver1 = "";
		private string lastReceiver2 = "";
		private bool initialized = false;
		private bool isPerson = false;
		
		public override void Init()
		{
			try
			{
				if (containingAtom.type == "Person") { isPerson = true; }
				else if (containingAtom.type == "Empty") { isPerson = false; }
				else
				{
					SuperController.LogMessage("SilverRoper: Must be attached to an atom of type Person or Empty.");
					this.enabled = false;
				}
				
				isAttached = NewToggle("isAttached", "Attached", false, false);
				
				atomJSON = new JSONStorableStringChooser("targetAtom1", SuperController.singleton.GetAtomUIDs(), null, "Atom", ReceiverChoices);
				RegisterStringChooser(atomJSON);				
				UIDynamicPopup dp = CreateFilterablePopup(atomJSON);
				AtomChocies();
				dp.popupPanelHeight = 700f;
				dp.popup.onOpenPopupHandlers += AtomChocies;

				receiverJSON = new JSONStorableStringChooser("targetReceiver1", null, null, "Connection", UpdateReceiverChoice);
				RegisterStringChooser(receiverJSON);
				dp = CreateFilterablePopup(receiverJSON);
				dp.popupPanelHeight = 700f;
				
				if (isPerson)
				{
					atom2JSON = new JSONStorableStringChooser("targetAtom2", SuperController.singleton.GetAtomUIDs(), null, "Atom 2", ReceiverChoices2);
					RegisterStringChooser(atom2JSON);
					dp = CreateFilterablePopup(atom2JSON);
					AtomChocies2();
					dp.popupPanelHeight = 700f;
					dp.popup.onOpenPopupHandlers += AtomChocies2;
					
					receiver2JSON = new JSONStorableStringChooser("targetReceiver2", null, null, "Connection 2", UpdateReceiverChoice);
					RegisterStringChooser(receiver2JSON);
					dp = CreateFilterablePopup(receiver2JSON);
					dp.popupPanelHeight = 700f;
				}
				
				ropeLength = NewSlider("Rope Length", 1f, 0.0001f, 5f, true);
				ropeRenderExtend = NewSlider("Rope Extend", 1f, 0.85f, 1.15f, true);
				ropeRadius = NewSlider("Rope Radius", 0.01f, 0.001f, 0.1f, true);
				
				CreateSpacer(false).height = 20;
				
				HSVColor c; c.H = 0f; c.S = 0f; c.V = 0f;
				ropeColor = new JSONStorableColor("Rope Color", c, UpdateColor);
				RegisterColor(ropeColor);
				colorPicker = CreateColorPicker(ropeColor, false);
				
				var btn = CreateButton("Set All Rope Colors", false);
                btn.button.onClick.AddListener(() => { SetAllRopeColors(); });
				
				CreateSpacer(false).height = 20;
				
				disableRot = NewToggle("disableRot", "Disable Target Rotation", true, false);
				
				CreateSpacer(true).height = 20;
				
				caOffsetX = NewSlider("Connection Offset 1 X", 0f, -0.3f, 0.3f, true);				
				caOffsetY = NewSlider("Connection Offset 1 Y", 0f, -0.3f, 0.3f, true);			
				caOffsetZ = NewSlider("Connection Offset 1 Z", 0f, -0.3f, 0.3f, true);
				
				if (isPerson)
				{
					CreateSpacer(true).height = 20;
					
					caOffset2X = NewSlider("Connection Offset 2 X", 0f, -0.3f, 0.3f, true);				
					caOffset2Y = NewSlider("Connection Offset 2 Y", 0f, -0.3f, 0.3f, true);				
					caOffset2Z = NewSlider("Connection Offset 2 Z", 0f, -0.3f, 0.3f, true);
				}
			}
			catch (Exception e)
			{
				SuperController.LogError("Exception caught: " + e);
			}
		}
		
		private void Start()
		{
			ropeLength.slider.onValueChanged.AddListener((value) => { UpdateRopeLength(); });
			ropeRenderExtend.slider.onValueChanged.AddListener((value) => { ropeLengthMulti = value; });
			ropeRadius.slider.onValueChanged.AddListener((value) => { ropeScale = value; });
			caOffsetX.slider.onValueChanged.AddListener((value) => { UpdateCAOffset(); });
			caOffsetY.slider.onValueChanged.AddListener((value) => { UpdateCAOffset(); });
			caOffsetZ.slider.onValueChanged.AddListener((value) => { UpdateCAOffset(); });
			if (isPerson)
			{
				caOffset2X.slider.onValueChanged.AddListener((value) => { UpdateAOffset(); });
				caOffset2Y.slider.onValueChanged.AddListener((value) => { UpdateAOffset(); });
				caOffset2Z.slider.onValueChanged.AddListener((value) => { UpdateAOffset(); });
			}
			
			UpdateRopeLength();
			ropeLengthMulti = ropeRenderExtend.val;
			ropeScale = ropeRadius.val;		
			lastReceiver1 = receiverJSON.val;
			if (isPerson)
			{
				lastReceiver2 = receiver2JSON.val;
			}
			
			if (isAttached.val && receiverJSON.val != null && receiverJSON.val != "" && receiverJSON.val != "None")
			{
				if (!isPerson)
				{
					Attach(true);
				}
				else if (receiver2JSON.val != null && receiver2JSON.val != "" && receiver2JSON.val != "None")
				{
					Attach(true);
				}
			}
			
			if (isPerson)
			{
				if (atomJSON.val == "" || atomJSON.val == "None" || atomJSON.val == null)
				{
					atomJSON.val = containingAtom.uid;
				}				
				if (atom2JSON.val == "" || atom2JSON.val == "None" || atom2JSON.val == null)
				{
					atom2JSON.val = containingAtom.uid;
				}
			}
			else if (atomJSON.choices.Count > 0 && (atomJSON.val == "" || atomJSON.val == "None" || atomJSON.val == null))
			{
				atomJSON.val = atomJSON.choices[0];
			}
			
			isAttached.toggle.onValueChanged.AddListener((value) =>
				{
					if (isAttached.val)
					{
						Attach();
					}
					else
					{
						Detach();
					}
				});
			
			initialized = true;
		}
		
		private void Update()
		{
			if (isAttached.val)
			{
				AlignRope();
			}
		}
		
		private void AtomChocies()
		{
			atomJSON.choices = SuperController.singleton.GetAtoms().Where(atom => atom.category == "People").Select(atom => atom.uid).ToList();
			
		}
		private void AtomChocies2()
		{
			if (!isPerson)
			{
				atom2JSON.choices = SuperController.singleton.GetAtoms().Where(atom => atom.category == "People").Select(atom => atom.uid).ToList();
			}
		}
		
		private void ReceiverChoices(string atomUID)
		{
			List<string> choices = new List<string>();			
			choices.Add("lHand");
			choices.Add("rHand");
			choices.Add("lForeArm");
			choices.Add("rForeArm");
			choices.Add("lShldr");
			choices.Add("rShldr");
			choices.Add("lFoot");
			choices.Add("rFoot");
			choices.Add("lShin");
			choices.Add("rShin");
			choices.Add("lThigh");
			choices.Add("rThigh");
			choices.Add("hip");
			choices.Add("chest");
			choices.Add("neck");
			choices.Add("abdomen");
			choices.Add("abdomen2");
			
			receiverJSON.choices = choices;
			receivingAtom = SuperController.singleton.GetAtomByUid(atomUID);
			if (receivingAtom != null)
			{
				
				controllers = new Dictionary<string, FreeControllerV3>();
				foreach (FreeControllerV3 fc in receivingAtom.freeControllers)
				{
					foreach (string str in choices)
					{
						if (fc.name == str + "Control")
						{
							controllers.Add(str, fc);
						}
					}
				}
			}
		}
		
		private void ReceiverChoices2(string atomUID)
		{
			if (!isPerson) { return; }
			
			List<string> choices = new List<string>();			
			choices.Add("lHand");
			choices.Add("rHand");
			choices.Add("lForeArm");
			choices.Add("rForeArm");
			choices.Add("lShldr");
			choices.Add("rShldr");
			choices.Add("lFoot");
			choices.Add("rFoot");
			choices.Add("lShin");
			choices.Add("rShin");
			choices.Add("lThigh");
			choices.Add("rThigh");
			choices.Add("hip");
			choices.Add("chest");
			choices.Add("neck");
			choices.Add("abdomen");
			choices.Add("abdomen2");
			
			receiver2JSON.choices = choices;			
			receivingAtom2 = SuperController.singleton.GetAtomByUid(atomUID);
			if (receivingAtom2 != null)
			{
				controllers2 = new Dictionary<string, FreeControllerV3>();
				foreach (FreeControllerV3 fc in receivingAtom2.freeControllers)
				{
					foreach (string str in choices)
					{
						if (fc.name == str + "Control")
						{
							controllers2.Add(str, fc);
						}
					}
				}
			}
		}
		
		private void UpdateReceiverChoice(string receiverName)
		{
			if (!initialized) { return; }
			if (isAttached.val && (lastReceiver1 != receiverJSON.val || (isPerson && lastReceiver2 != receiver2JSON.val )))
			{
				isAttached.val = false;
			}
		}
		
		private void UpdateColor(JSONStorableColor c)
		{
			if (ropeMat)
			{
				ropeMat.color = HSVColorPicker.HSVToRGB(c.val);
			}
		}
		
		private void UpdateRopeLength()
		{
			if (!cj) { return; }
			SoftJointLimit sjl = new SoftJointLimit();
			sjl.limit = ropeLength.val;
			cj.linearLimit = sjl;
		}
		
		private void Attach(bool isStart=false)
        {
			ropeScale = ropeRadius.val;			
			receivingAtom = SuperController.singleton.GetAtomByUid(atomJSON.val);
			if (!receivingAtom) { SuperController.LogMessage("SilverRoper: Atom is null"); return; }
			if (isPerson)
			{
				receivingAtom2 = SuperController.singleton.GetAtomByUid(atom2JSON.val);
				if (!receivingAtom2) { SuperController.LogMessage("SilverRoper: Atom 2 is null"); return; }
			}
			
			fixedOffset = Vector3.zero;
			Rigidbody refBody;
			if (receiverJSON.val == "lHand")
			{
				refBody = receivingAtom.rigidbodies.First(fc => fc.name == "lHand");
				targetBody = receivingAtom.rigidbodies.First(fc => fc.name == "lForeArm");
				fixedOffset.x = -Vector3.Distance(refBody.transform.position, targetBody.transform.position);
				fixedOffset.y = 0.03f;
			}
			else if (receiverJSON.val == "rHand")
			{
				refBody = receivingAtom.rigidbodies.First(fc => fc.name == "rHand");
				targetBody = receivingAtom.rigidbodies.First(fc => fc.name == "rForeArm");
				fixedOffset.x = Vector3.Distance(refBody.transform.position, targetBody.transform.position);
				fixedOffset.y = 0.03f;
			}
			else if (receiverJSON.val == "lFoot")
			{
				refBody = receivingAtom.rigidbodies.First(fc => fc.name == "lFoot");
				targetBody = receivingAtom.rigidbodies.First(fc => fc.name == "lShin");
				fixedOffset.y = -Vector3.Distance(refBody.transform.position, targetBody.transform.position);
			}
			else if (receiverJSON.val == "rFoot")
			{
				refBody = receivingAtom.rigidbodies.First(fc => fc.name == "rFoot");
				targetBody = receivingAtom.rigidbodies.First(fc => fc.name == "rShin");
				fixedOffset.y = -Vector3.Distance(refBody.transform.position, targetBody.transform.position);
			}
			else
			{
				refBody = targetBody = receivingAtom.rigidbodies.First(fc => fc.name == receiverJSON.val);
				fixedOffset = Vector3.zero;
			}
			
			if (!targetBody) { SuperController.LogMessage("SilverRoper: Target is null"); return; }
			lastReceiver1 = receiverJSON.val;
			
			targetController = null;
			controllers.TryGetValue(receiverJSON.val, out targetController);			
			if (targetController)
			{
				rState = targetController.currentRotationState;
				pState = targetController.currentPositionState;
				startPos = targetController.transform.position;
				startRot = targetController.transform.rotation;
			}
			
			if (isPerson)
			{
				fixedOffset2 = Vector3.zero;
				if (receiver2JSON.val == "lHand")
				{
					refBody = receivingAtom2.rigidbodies.First(fc => fc.name == "lHand");
					targetBody2 = receivingAtom2.rigidbodies.First(fc => fc.name == "lForeArm");
					fixedOffset2.x = -Vector3.Distance(refBody.transform.position, targetBody2.transform.position);
					fixedOffset2.y = 0.03f;
				}
				else if (receiver2JSON.val == "rHand")
				{
					refBody = receivingAtom2.rigidbodies.First(fc => fc.name == "rHand");
					targetBody2 = receivingAtom2.rigidbodies.First(fc => fc.name == "rForeArm");
					fixedOffset2.x = Vector3.Distance(refBody.transform.position, targetBody2.transform.position);
					fixedOffset2.y = 0.03f;
				}
				else if (receiver2JSON.val == "lFoot")
				{
					refBody = receivingAtom2.rigidbodies.First(fc => fc.name == "lFoot");
					targetBody2 = receivingAtom2.rigidbodies.First(fc => fc.name == "lShin");
					fixedOffset2.y = -Vector3.Distance(refBody.transform.position, targetBody2.transform.position);
				}
				else if (receiver2JSON.val == "rFoot")
				{
					refBody = receivingAtom2.rigidbodies.First(fc => fc.name == "rFoot");
					targetBody2 = receivingAtom2.rigidbodies.First(fc => fc.name == "rShin");
					fixedOffset2.y = -Vector3.Distance(refBody.transform.position, targetBody2.transform.position);
				}
				else
				{
					refBody = targetBody2 = receivingAtom2.rigidbodies.First(fc => fc.name == receiver2JSON.val);
					fixedOffset2 = Vector3.zero;
				}
				
				if (!targetBody2) { SuperController.LogMessage("SilverRoper: Target 2 is null"); return; }
				lastReceiver2 = receiver2JSON.val;
				
				targetController2 = null;
				controllers2.TryGetValue(receiver2JSON.val, out targetController2);
				if (targetController2)
				{
					rState2 = targetController.currentRotationState;
					pState2 = targetController.currentPositionState;
					startPos2 = targetController2.transform.position;
					startRot2 = targetController2.transform.rotation;
				}
			}
			
			if (isPerson)
			{
				cj = targetBody2.gameObject.GetComponents<ConfigurableJoint>().Where(j => j.connectedBody == null || j.connectedBody == targetBody).FirstOrDefault();
				if (!cj)
				{
					cj = targetBody2.gameObject.AddComponent<ConfigurableJoint>();
				}
			}
			else
			{
				Rigidbody controlRB = containingAtom.mainController.followWhenOffRB;
				containingAtom.mainController.RBMass = targetBody.mass * 10f;
				containingAtom.mainController.RBDrag = 0.25f;
				containingAtom.mainController.RBAngularDrag = 0.25f;
				cj = controlRB.GetComponents<ConfigurableJoint>().Where(j => j.connectedBody == null || j.connectedBody == targetBody).FirstOrDefault();
				if (!cj)
				{
					cj = controlRB.gameObject.AddComponent<ConfigurableJoint>();
				}
			}
			
			cj.connectedBody = null;
			cj.xMotion = ConfigurableJointMotion.Limited;
			cj.yMotion = ConfigurableJointMotion.Limited;
			cj.zMotion = ConfigurableJointMotion.Limited;
			
			SoftJointLimit sjl = new SoftJointLimit();
			if (isStart)
			{
				sjl.limit = ropeLength.val;
			}
			else
			{
				sjl.limit = Vector3.Distance(cj.transform.position, targetBody.transform.position);
				ropeLength.val = sjl.limit;
			}
			cj.linearLimit = sjl;
			
			SoftJointLimitSpring sjls = new SoftJointLimitSpring();
			sjls.spring = 400f;
			sjls.damper = 5f;
			cj.linearLimitSpring = sjls;
			
			JointDrive jd = new JointDrive();
			jd.positionSpring = 400f;
			jd.positionDamper = 5f;
			cj.xDrive = jd;
			cj.yDrive = jd;
			cj.zDrive = jd;
			
			jd.positionSpring = 500f;
			cj.slerpDrive = jd;
			
			if (isPerson)
			{
				cj.anchor = new Vector3(caOffset2X.val, caOffset2Y.val, caOffset2Z.val) + fixedOffset2;
			}
			else
			{
				cj.connectedMassScale = 10f;
			}
			cj.autoConfigureConnectedAnchor = false;
			cj.connectedBody = targetBody;
			cj.connectedAnchor = new Vector3(caOffsetX.val, caOffsetY.val, caOffsetZ.val)+ fixedOffset;
			
			if (targetController != null)
			{
				if (disableRot.val)
				{
					targetController.currentRotationState = FreeControllerV3.RotationState.Off;
				}
				targetController.currentPositionState = FreeControllerV3.PositionState.Off;
			}
			
			if (isPerson && targetController2 != null)
			{
				if (disableRot.val)
				{
					targetController2.currentRotationState = FreeControllerV3.RotationState.Off;
				}
				targetController2.currentPositionState = FreeControllerV3.PositionState.Off;
			}
			
			CreateRope();			
			AlignRope();
        }
		
		private void UpdateCAOffset()
		{
			if (!cj) { return; }
			cj.connectedAnchor = new Vector3(caOffsetX.val, caOffsetY.val, caOffsetZ.val) + fixedOffset;
		}
		
		private void UpdateAOffset()
		{
			if (!isPerson || !cj) { return; }
			cj.anchor = new Vector3(caOffset2X.val, caOffset2Y.val, caOffset2Z.val) + fixedOffset2;
		}
		
		private void CreateRope()
		{
			if (!rope)
			{
				rope = GameObject.CreatePrimitive(PrimitiveType.Cylinder);
			}
			
			Collider col = rope.GetComponent<Collider>();
			if (col)
			{
				col.enabled = false;
			}
			
			Renderer rend = rope.GetComponent<Renderer>();
			ropeMat = rend.material;
			ropeMat.color = HSVColorPicker.HSVToRGB(ropeColor.val);
		}
		
		private void DestroyRope()
		{
			if (rope)
			{
				GameObject.Destroy(rope);
				ropeMat = null;
			}
		}
		
		private void AlignRope()
		{
			Vector3 p1;
			if (isPerson)
			{
				p1 = targetBody2.transform.position + targetBody2.transform.TransformVector(cj.anchor);
			}
			else
			{
				p1 = cj.transform.position;
			}
			
			Vector3 p2 = targetBody.transform.position + targetBody.transform.TransformVector(cj.connectedAnchor);
			Vector3 localScale = new Vector3(ropeScale, Vector3.Distance(p1, p2) * 0.5f * ropeLengthMulti, ropeScale);
			rope.transform.localScale = localScale;
			Vector3 dir = p2 - p1;
			rope.transform.position = p1 + (dir.normalized * (dir.magnitude * 0.5f));
			rope.transform.rotation = Quaternion.FromToRotation (Vector3.up, dir);
		}
		
		private void Detach()
		{
			if (targetController != null)
			{
				targetController.currentRotationState = rState;
				targetController.currentPositionState = pState;
				
			}
			if (isPerson && targetController2 != null)
			{
				targetController2.currentRotationState = rState2;
				targetController2.currentPositionState = pState2;
			}
			if (cj)
			{
				cj.connectedBody = null;
				GameObject.Destroy(cj);
			}
			
			DestroyRope();
		}
		
		private JSONStorableFloat NewSlider(string name, float val, float min, float max, bool right)
		{
			JSONStorableFloat storable = new JSONStorableFloat(name, val, min, max, true, true);
			CreateSlider(storable, right); RegisterFloat(storable); return storable;
		}
		
		private JSONStorableBool NewToggle(string name, string label, bool val, bool right)
		{
			JSONStorableBool storable = new JSONStorableBool(name, val);
			var t = CreateToggle(storable, right); RegisterBool(storable); t.label = label;
			return storable;
		}
		
		public void SetAllRopeColors()
		{
			foreach (Atom a in SuperController.singleton.GetAtoms())
			{
				List<string> ids = a.GetStorableIDs();
				foreach (string id in ids)
				{
					if (id.StartsWith("plugin#") && id.Contains("ClockwiseSilver.SilverRoper"))
					{
						JSONStorable storable = a.GetStorableByID(id);
						JSONStorableColor c = storable.GetColorJSONParam("Rope Color");
						c.val = ropeColor.val;
					}
				}
			}
		}
		
		private void OnDisable()
		{
			Detach();
			DestroyRope();
			if (cj)
			{
				GameObject.Destroy(cj);
			}
		}
	}
}